# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   lun 14 nov 2016 23:47:27 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016 Lele Gaifax
#

import sys

from sqlalchemy import (create_engine, Column, ForeignKey, ForeignKeyConstraint,
                        Integer, Sequence, String)
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker


Base = declarative_base()

class Country(Base):
    __tablename__ = 'countries'

    id = Column(String(2), primary_key=True, nullable=False)
    name = Column(String(64), nullable=False)


class City(Base):
    __tablename__ = 'cities'

    country_id = Column(String(2), ForeignKey('countries.id'),
                        primary_key=True, nullable=False)
    code = Column(String(4), primary_key=True, nullable=False)
    name = Column(String(64), nullable=False)

    country = relationship('Country')

class Address(Base):
    __tablename__ = 'addresses'
    __table_args__ = (
        ForeignKeyConstraint(['country_id', 'city_code'],
                             ['cities.country_id', 'cities.code']),
    )

    id = Column(Integer, Sequence('address_id_seq'), primary_key=True)
    country_id = Column(String(2), ForeignKey('countries.id'),
                        nullable=True)
    city_code = Column(String(4), nullable=True)
    street = Column(String(64), nullable=False)

    country = relationship('Country')
    city = relationship('City',
                        primaryjoin=
                        "and_(City.code == foreign(Address.city_code),"
                        "City.country_id == Address.country_id)")

if len(sys.argv) == 2:
    url = 'sqlite:////tmp/testdbloady.sqlite'
    e = create_engine(url)

    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    elif sys.argv[1] == 'test':
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        addresses = session.query(Address).order_by(Address.city_code).all()
        assert len(addresses) == 2



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Adaptor test
# :Created:   sab 09 gen 2016 11:11:05 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

from sqlalchemy import create_engine
from sqlalchemy.orm import sessionmaker
from metapensiero.sqlalchemy.dbloady.load import load
from model import Base, Person, Result


def tweak_data(cls, key, data):
    if cls is Person:
        data['username'] = data['firstname'][:1] + data['lastname']
    elif cls is Result:
        data['vote'] = data['vote'] * 2
    return data


def main():
    e = create_engine('sqlite:///:memory:')
    Base.metadata.create_all(e)

    smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
    session = smaker()

    load('data.yaml', session, adaptor=tweak_data)

    jdoe = session.query(Person).filter(Person.username == 'JDoe').first()

    assert jdoe.firstname == 'John'
    assert len(jdoe.results) == 1
    assert jdoe.results[0].vote == 20


if __name__ == '__main__':
    main()



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- YAML based data dumper
# :Created:   dom 06 mar 2016 17:48:31 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

from logging import getLogger
from inspect import isgenerator
from io import open
from itertools import chain
from os.path import isabs
import sys

import pkg_resources

from sqlalchemy import create_engine, inspect
from sqlalchemy.exc import NoInspectionAvailable
from sqlalchemy.orm import sessionmaker
from sqlalchemy.orm.collections import InstrumentedList

from ruamel import yaml

from . import resolve_class_name
from .entity import Entity


OK, SPEC_FILE_ERROR, PRELOAD_EXCEPTION = range(3)

logger = getLogger(__name__)

if sys.version_info.major >= 3:
    basestring = str


def dump(pivots, specs):
    """Dump database content into a YAML document.

    :param pivots: either a single instance or an iterable of them
    :param specs: an sequence of *spec entries*, each describing how a particular entity shall
                  be serialized
    :rtype: list
    :returns: a sequence of items, one for each entity declared in `specs` (and in that order),
              containing the data of the collected instances of that entity

    A *spec entry* is a dictionary containing at least an ``entity`` slot with the fully
    qualified dotted name of a SQLAlchemy mapped class and a ``key`` slot with either a single
    string or a list of strings, the name of the attributes that uniquely identity a single
    instance. Remaining fields may be listed in the ``other`` slot.
    """

    specs_by_qualname = {spec['entity']: spec for spec in specs}

    seen = set()
    entities = {}
    instances = {}

    def get_identity(instance):
        klass = instance.__class__
        qualname = instance.__module__ + '.' + klass.__name__
        if qualname not in specs_by_qualname:
            return

        spec = specs_by_qualname[qualname]

        key = spec['key']
        if callable(key):
            key = key(instance)
        if isinstance(key, str):
            identity = (qualname, getattr(instance, key))
        else:
            assert isinstance(key, (tuple, list))
            identity = (qualname, tuple(getattr(instance, f) for f in key))

        return identity

    def visit(instance):
        identity = get_identity(instance)
        if identity is None:
            return

        if identity in seen:
            return instances[identity]

        seen.add(identity)

        qualname = identity[0]
        spec = specs_by_qualname[qualname]
        key = spec['key']

        if qualname in entities:
            entity = entities[qualname]
        else:
            entity = entities[qualname] = {'entity': qualname, 'key': key, 'rows': []}

        if 'fields' in spec:
            fields = entity['fields'] = spec['fields']
            if isinstance(fields, basestring):
                attributes = [fields]
            else:
                attributes = fields

            data = instances[identity] = []
        else:
            fields = None
            attributes = []

            if isinstance(key, str):
                attributes.append(key)
            else:
                attributes.extend(key)

            other = spec.get('other', [])
            if isinstance(other, basestring):
                attributes.append(other)
            else:
                attributes.extend(other)

            data = instances[identity] = {}

        entity['rows'].append(data)

        for attr in attributes:
            value = getattr(instance, attr)

            if value is None:
                continue

            if isinstance(value, InstrumentedList):
                for i in value:
                    visit(i)
            else:
                try:
                    inspector = inspect(value)
                except NoInspectionAvailable:
                    if fields is None:
                        data[attr] = value
                    else:
                        data.append(value)
                else:
                    if inspector.is_instance:
                        value = visit(value)
                        if value is not None:
                            if fields is None:
                                data[attr] = value
                            else:
                                data.append(value)

        return data

    if isinstance(pivots, (list, tuple)) or isgenerator(pivots):
        for pivot in pivots:
            visit(pivot)
    else:
        visit(pivots)

    return [entities[spec['entity']] for spec in specs
            if spec['entity'] in entities]


def workhorse(uri, echo, preload, specfname, outputstream):
    engine = create_engine(uri, echo=echo)
    salogger = getattr(engine.logger, 'logger', None)
    if salogger is not None:
        for h in salogger.handlers:
            salogger.removeHandler(h)
    smaker = sessionmaker(autoflush=False, autocommit=False, bind=engine)

    session = smaker()

    if preload is not None:
        try:
            f = open(preload)
        except IOError:
            logger.error("Could not open preload script %r!" % preload)
            return PRELOAD_EXCEPTION

        logger.debug('Executing preload script %r...', preload)
        context = dict(session=session)
        try:
            code = compile(f.read(), preload, 'exec')
            exec(code, context)
        except Exception:
            logger.exception("Failure executing the preload script!")
            return PRELOAD_EXCEPTION
        finally:
            f.close()


    idmap = {}
    pivots = None

    with open(specfname, 'r', encoding='utf-8') as stream:
        loader = yaml.Loader(stream)
        entries = loader.get_data()
        if entries is None:
            logger.error('Could not read pivot entries from spec file')
            return SPEC_FILE_ERROR
        specs = loader.get_data()
        if specs is None:
            logger.error('Invalid spec file, should contain two documents')
            return SPEC_FILE_ERROR

    for entry in entries:
        model = resolve_class_name(entry['entity'])
        if 'rows' in entry:
            data = entry['rows']
        elif 'data' in entry:
            data = entry['data']
        else:
            data = None

        if data is not None:
            entity = Entity(model, entry['key'],
                            fields=entry.get('fields', None),
                            data=data,
                            loadonly=True)
            instances = entity(session, idmap)
        else:
            instances = iter(session.query(model))

        if pivots is None:
            pivots = instances
        else:
            pivots = chain(pivots, instances)


    result = dump(pivots, specs)

    yaml.dump(result, outputstream, default_flow_style=False)

    return OK


def path_spec(ps):
    if isabs(ps) or not ':' in ps:
        return ps
    pkgname, subpath = ps.split(':', 1)
    return pkg_resources.resource_filename(pkgname, subpath)


def main():
    import locale, logging
    from argparse import ArgumentParser, RawDescriptionHelpFormatter

    locale.setlocale(locale.LC_ALL, '')

    parser = ArgumentParser(
        description="Dump YAML representation of database content.",
        epilog=__doc__, formatter_class=RawDescriptionHelpFormatter)

    parser.add_argument("specfile", type=path_spec,
                        help="The YAML data file to containing dump specifications."
                        " It may be either a plain file name or a package"
                        " relative path like “package.name:some/file”.")
    parser.add_argument("output", default="-", nargs='?',
                        help="Where to write the YAML dump: “-” means stdout,"
                        " the default.")
    parser.add_argument("-u", "--sqlalchemy-uri", type=str, metavar="URI",
                        help="Specify the SQLAlchemy URI.", default=None)
    parser.add_argument("-p", "--preload", type=path_spec, metavar='SCRIPT',
                        help="Execute the given Python script before loading the"
                        " data files. It may be either a plain file name or a package"
                        " relative path like “package.name:some/file”.")
    parser.add_argument("-e", "--echo", default=False, action="store_true",
                        help="Activate SA engine echo")
    parser.add_argument("-d", "--debug", default=False, action="store_true",
                        help="Activate debug logging")

    args = parser.parse_args()

    logging.basicConfig(format='%(message)s',
                        level=logging.DEBUG if args.debug else logging.INFO)

    if args.output == '-':
        output = sys.stdout
    else:
        output = open(args.output, 'w', encoding='utf-8')

    return workhorse(args.sqlalchemy_uri, args.echo, args.preload, args.specfile, output)


if __name__ == '__main__':
    sys.exit(main())



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   ven 08 gen 2016 12:27:26 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

import sys

from sqlalchemy import create_engine, Column, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship, sessionmaker


Base = declarative_base()

class Person(Base):
    __tablename__ = 'persons'

    id = Column(Integer, primary_key=True)
    firstname = Column(String(64))
    lastname = Column(String(64))


class Exam(Base):
    __tablename__ = 'exams'

    id = Column(Integer, primary_key=True)
    description = Column(String(64))


class Result(Base):
    __tablename__ = 'results'

    id = Column(Integer, primary_key=True)
    idperson = Column(Integer,
                      ForeignKey('persons.id', ondelete='CASCADE'),
                      nullable=False)
    idexam = Column(Integer, ForeignKey('exams.id', ondelete='CASCADE'),
                    nullable=False)
    vote = Column(Integer)

    person = relationship('Person')
    exam = relationship('Exam')


if len(sys.argv) == 2:
    url = 'sqlite:////tmp/testdbloady.sqlite'
    e = create_engine(url)

    if sys.argv[1] == 'setup':
        Base.metadata.create_all(e)
    elif sys.argv[1] == 'test':
        smaker = sessionmaker(autoflush=False, autocommit=False, bind=e)
        session = smaker()

        all_results = session.query(Result).order_by(Result.vote).all()
        assert len(all_results) == 2
        assert all_results[0].vote == 5
        assert all_results[0].person.lastname == 'Foo'
        assert all_results[1].vote == 10
        assert all_results[1].person.lastname == 'Doe'



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Sample postload script
# :Created:   gio 14 gen 2016 11:36:58 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

import datetime
import model

event = session.query(model.Event).first()

assert event.starttime == datetime.time(12, 23)
assert event.endtime == datetime.time(15, 34)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- Test model
# :Created:   sab 09 gen 2016 10:55:39 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

from sqlalchemy import Column, ForeignKey, Integer, String
from sqlalchemy.ext.declarative import declarative_base
from sqlalchemy.orm import relationship


Base = declarative_base()

class Person(Base):
    __tablename__ = 'persons'

    username = Column(String(64), primary_key=True)
    firstname = Column(String(64))
    lastname = Column(String(64))


class Exam(Base):
    __tablename__ = 'exams'

    id = Column(Integer, primary_key=True)
    description = Column(String(64))


class Result(Base):
    __tablename__ = 'results'

    id = Column(Integer, primary_key=True)
    username = Column(Integer,
                      ForeignKey('persons.username', ondelete='CASCADE'),
                      nullable=False)
    idexam = Column(Integer, ForeignKey('exams.id', ondelete='CASCADE'),
                    nullable=False)
    vote = Column(Integer)

    person = relationship('Person', backref='results')
    exam = relationship('Exam')



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- YAML based data loader
# :Created:   mer 10 feb 2010 14:35:05 CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2010-2017 Lele Gaifax
#

from os.path import join, normpath

from ruamel import yaml


def resolve_class_name(classname):
    """Import a particular Python class given its full dotted name.

    :param classname: full dotted name of the class,
                      such as "package.module.ClassName"
    :rtype: the Python class
    """

    modulename, _, classname = classname.rpartition('.')
    module = __import__(modulename, fromlist=[classname])
    return getattr(module, classname)


class File(yaml.YAMLObject):
    """Facility to read the content of an external file.

    The value of field may be loaded from an external file, given its pathname which is
    interpreted as relative to the position of the YAML file currently loading::

        - entity: cpi.models.Document
          key: filename
          data:
            - filename: image.gif
              content: !File {path: ../image.gif}

    Alternatively, the content may be inline, possibly compressed::

        - entity: cpi.models.Document
          key: filename
          data:
            - filename: image.gif
              content: !File
                compressor: lzma
                content: !!binary |
                  /Td6WFoAAATm1rRGAgAhA...
    """

    yaml_tag = '!File'

    basedir = None

    def __init__(self, path=None, content=None, compressor=None):
        self.path = path
        self.content = content
        self.compressor = compressor

    def read(self):
        # PyYAML does not execute the __init__ method
        path = getattr(self, 'path', None)
        if path is None:
            content = getattr(self, 'content', None)
            if content is None:
                raise RuntimeError('The !File object requires either a "path"'
                                   ' or a "content" argument')
            compressor = getattr(self, 'compressor', None)
            if compressor is not None:
                if compressor == 'lzma':
                    from lzma import decompress
                elif compressor == 'gzip':
                    from gzip import decompress
                else:
                    raise RuntimeError('Unsupported compressor: %r' % compressor)
                content = decompress(content)
        else:
            fullpath = normpath(join(self.basedir, path))
            content = open(fullpath, 'rb').read()
        return content


class SQL(yaml.YAMLObject):
    """Raw SQL statement."""

    yaml_tag = '!SQL'

    def __init__(self, query, params=None):
        self.query = query
        self.params = params

    def fetch(self, session):
        # PyYAML does not execute the __init__ method
        try:
            query = self.query
        except AttributeError:
            raise RuntimeError('The !SQL object requires a "query" argument')
        try:
            params = self.params
        except AttributeError:
            params = None
        return session.scalar(query, params)



# -*- coding: utf-8 -*-
# :Project:   metapensiero.sqlalchemy.dbloady -- YAML based data loader
# :Created:   ven  1 gen 2016, 16.19.54, CET
# :Author:    Lele Gaifax <lele@metapensiero.it>
# :License:   GNU General Public License version 3 or later
# :Copyright: © 2016, 2017 Lele Gaifax
#

from io import open
import os

from setuptools import setup, find_packages

here = os.path.abspath(os.path.dirname(__file__))
with open(os.path.join(here, 'README.rst'), encoding='utf-8') as f:
    README = f.read()
with open(os.path.join(here, 'CHANGES.rst'), encoding='utf-8') as f:
    CHANGES = f.read()
with open(os.path.join(here, 'version.txt'), encoding='utf-8') as f:
    VERSION = f.read().strip()

setup(
    name="metapensiero.sqlalchemy.dbloady",
    version=VERSION,
    url="https://bitbucket.org/lele/metapensiero.sqlalchemy.dbloady",

    description="YAML based data loader",
    long_description=README + u'\n\n' + CHANGES,

    author="Lele Gaifax",
    author_email="lele@metapensiero.it",

    license="GPLv3+",
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python",
        "Programming Language :: Python :: 2",
        "Programming Language :: Python :: 2.7",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.4",
        "Programming Language :: Python :: 3.5",
        "Programming Language :: Python :: 3.6",
        "Topic :: Utilities",
    ],
    keywords='',

    packages=find_packages('src'),
    package_dir={'': 'src'},
    namespace_packages=['metapensiero', 'metapensiero.sqlalchemy'],

    install_requires=[
        'progressbar2',
        'ruamel.yaml',
        'setuptools',
        'sqlalchemy',
    ],
    extras_require={
        'dev': [
            'metapensiero.tool.bump_version',
            'readme_renderer',
        ]
    },

    entry_points="""\
    [console_scripts]
    dbdumpy = metapensiero.sqlalchemy.dbloady.dump:main
    dbloady = metapensiero.sqlalchemy.dbloady.load:main
    """,
)
